@@ -215,7 +215,13 @@ static void config_parse_keys(GKeyFile *key_file, char *group)
     config_get_string(key_file, group, "log-file", &option.log_filename, NULL);
     config_get_bool(key_file, group, "log-append", &option.log_append);
     config_get_bool(key_file, group, "log-strip", &option.log_strip);
-    config_get_string(key_file, group, "map", &option.map, NULL);
+    config_get_string(key_file, group, "map", &string, NULL);
+    if (string != NULL)
+    {
+        option_parse_mappings(string);
+        g_free((void *)string);
+        string = NULL;
+    }
     config_get_string(key_file, group, "color", &string, NULL);
     if (string != NULL)
     {
@@ -402,13 +408,12 @@ static char *match_and_replace(const char *str, const char *pattern, char *devic
     assert(pattern != NULL);
     assert(device != NULL);
 
-    char *string = malloc(strlen(device) + PATH_MAX);
+    char *string = strndup(device, PATH_MAX);
     if (string == NULL)
     {
         tio_debug_printf("Failure allocating string memory\n");
         return NULL;
     }
-    strcpy(string, device);
 
     /* Find matches of pattern in str. For each match, replace any '%mN' in the
      * copy of the device string with the corresponding match subexpression and
