@@ -64,6 +64,7 @@
 #include "script.h"
 #include "xymodem.h"
 #include "fs.h"
+#include "readline.h"
 
 /* tty device listing configuration */
 
@@ -87,9 +88,6 @@
 #define CMSPAR   010000000000
 #endif
 
-#define MAX_LINE_LEN PATH_MAX
-#define MAX_HISTORY 1000
-
 #define KEY_0 0x30
 #define KEY_1 0x31
 #define KEY_2 0x32
@@ -173,7 +171,7 @@ static char *tty_buffer_write_ptr = tty_buffer;
 static pthread_t thread;
 static int pipefd[2];
 static pthread_mutex_t mutex_input_ready = PTHREAD_MUTEX_INITIALIZER;
-static char line[MAX_LINE_LEN];
+static char line[PATH_MAX];
 
 static void optional_local_echo(char c)
 {
@@ -581,7 +579,7 @@ static int tio_readln(void)
     char *p = line;
 
     /* Read line, accept BS and DEL as rubout characters */
-    for (p = line ; p < &line[MAX_LINE_LEN-1]; )
+    for (p = line ; p < &line[PATH_MAX-1]; )
     {
         if (read(pipefd[0], p, 1) > 0)
         {
@@ -625,20 +623,21 @@ static void mappings_print(void)
     if (option.map_i_cr_nl || option.map_ign_cr || option.map_i_ff_escc ||
         option.map_i_nl_cr || option.map_i_nl_crnl || option.map_o_cr_nl ||
         option.map_o_del_bs || option.map_o_nl_crnl || option.map_o_ltu ||
-        option.map_o_nulbrk || option.map_o_msblsb)
+        option.map_o_nulbrk || option.map_i_msb2lsb || option.map_o_ign_cr)
     {
-        tio_printf(" Mappings:%s%s%s%s%s%s%s%s%s%s%s",
+        tio_printf(" Mappings:%s%s%s%s%s%s%s%s%s%s%s%s",
                 option.map_i_cr_nl ? " ICRNL" : "",
                 option.map_ign_cr ? " IGNCR" : "",
                 option.map_i_ff_escc ? " IFFESCC" : "",
                 option.map_i_nl_cr ? " INLCR" : "",
                 option.map_i_nl_crnl ? " INLCRNL" : "",
+                option.map_i_msb2lsb ? " IMSB2LSB" : "",
                 option.map_o_cr_nl ? " OCRNL" : "",
                 option.map_o_del_bs ? " ODELBS" : "",
                 option.map_o_nl_crnl ? " ONLCRNL" : "",
                 option.map_o_ltu ? " OLTU" : "",
                 option.map_o_nulbrk ? " ONULBRK" : "",
-                option.map_o_msblsb ? " MSB2LSB" : "");
+                option.map_o_ign_cr ? " OIGNCR" : "");
     }
     else
     {
@@ -773,28 +772,32 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
                         tio_printf("INLCRNL is %s", option.map_i_nl_crnl ? "set" : "unset");
                         break;
                     case KEY_5:
+                        option.map_i_msb2lsb = !option.map_i_msb2lsb;
+                        tio_printf("IMSB2LSB is %s", option.map_i_msb2lsb ? "set" : "unset");
+                        break;
+                    case KEY_6:
                         option.map_o_cr_nl = !option.map_o_cr_nl;
                         tio_printf("OCRNL is %s", option.map_o_cr_nl ? "set" : "unset");
                         break;
-                    case KEY_6:
+                    case KEY_7:
                         option.map_o_del_bs = !option.map_o_del_bs;
                         tio_printf("ODELBS is %s", option.map_o_del_bs ? "set" : "unset");
                         break;
-                    case KEY_7:
+                    case KEY_8:
                         option.map_o_nl_crnl = !option.map_o_nl_crnl;
                         tio_printf("ONLCRNL is %s", option.map_o_nl_crnl ? "set" : "unset");
                         break;
-                    case KEY_8:
+                    case KEY_9:
                         option.map_o_ltu = !option.map_o_ltu;
                         tio_printf("OLTU is %s", option.map_o_ltu ? "set" : "unset");
                         break;
-                    case KEY_9:
+                    case KEY_A:
                         option.map_o_nulbrk = !option.map_o_nulbrk;
                         tio_printf("ONULBRK is %s", option.map_o_nulbrk ? "set" : "unset");
                         break;
-                    case KEY_A:
-                        option.map_o_msblsb = !option.map_o_msblsb;
-                        tio_printf("MSB2LSB is %s", option.map_o_msblsb ? "set" : "unset");
+                    case KEY_B:
+                        option.map_o_ign_cr = !option.map_o_ign_cr;
+                        tio_printf("OIGNCR is %s", option.map_o_ign_cr ? "set" : "unset");
                         break;
                     default:
                         tio_error_print("Invalid input");
@@ -994,18 +997,20 @@ void handle_command_sequence(char input_char, char *output_char, bool *forward)
                         option.map_i_nl_cr ? "Unset" : "Set");
                 tio_printf(" (4) INLCRNL: %s mapping NL to CR-NL on input",
                         option.map_i_nl_cr ? "Unset" : "Set");
-                tio_printf(" (5) OCRNL: %s mapping CR to NL on output",
+                tio_printf(" (5) IMSB2LSB: %s mapping MSB bit order to LSB on input",
+                        option.map_i_msb2lsb ? "Unset" : "Set");
+                tio_printf(" (6) OCRNL: %s mapping CR to NL on output",
                         option.map_o_cr_nl ? "Unset" : "Set");
-                tio_printf(" (6) ODELBS: %s mapping DEL to BS on output",
+                tio_printf(" (7) ODELBS: %s mapping DEL to BS on output",
                         option.map_o_del_bs ? "Unset" : "Set");
-                tio_printf(" (7) ONLCRNL: %s mapping NL to CR-NL on output",
+                tio_printf(" (8) ONLCRNL: %s mapping NL to CR-NL on output",
                         option.map_o_nl_crnl ? "Unset" : "Set");
-                tio_printf(" (8) OLTU: %s mapping lowercase to uppercase on output",
+                tio_printf(" (9) OLTU: %s mapping lowercase to uppercase on output",
                         option.map_o_ltu ? "Unset" : "Set");
-                tio_printf(" (9) ONULBRK: %s mapping NUL to send break signal on output",
+                tio_printf(" (a) ONULBRK: %s mapping NUL to send break signal on output",
                         option.map_o_nulbrk ? "Unset" : "Set");
-                tio_printf(" (a) MSB2LSB: %s mapping MSB bit order to LSB on output",
-                        option.map_o_msblsb ? "Unset" : "Set");
+                tio_printf(" (b) OIGNCR: %s ignoring CR on output",
+                        option.map_o_ign_cr ? "Unset" : "Set");
 
                 // Process next input character as sub command
                 sub_command = SUBCOMMAND_MAP;
@@ -2135,6 +2140,10 @@ void forward_to_tty(int fd, char output_char)
     {
         output_char = '\n';
     }
+    if ((output_char == '\r') && (option.map_o_ign_cr))
+    {
+        return;
+    }
 
     /* Map newline character */
     if ((output_char == '\n' || output_char == '\r') && (option.map_o_nl_crnl))
@@ -2207,17 +2216,6 @@ void forward_to_tty(int fd, char output_char)
     }
 }
 
-static void print_line(const char *string, int cursor_pos)
-{
-    clear_line();
-    print("%s", string);
-    print("\r"); // Move the cursor back to the beginning
-    for (int i = 0; i < cursor_pos; ++i)
-    {
-        print("\x1b[C"); // Move the cursor right
-    }
-}
-
 int tty_connect(void)
 {
     fd_set rdfs;           /* Read file descriptor set */
@@ -2372,18 +2370,7 @@ int tty_connect(void)
     }
 
     // Initialize readline like history
-    char *history[MAX_HISTORY];
-    int history_count = 0;
-    int history_index = 0;
-
-    for (int i = 0; i < MAX_HISTORY; ++i)
-    {
-        history[i] = NULL;
-    }
-
-    int line_length = 0;
-    int cursor_pos = 0;
-    int escape = 0;
+    readline_init();
 
     /* Input loop */
     while (true)
@@ -2520,13 +2507,13 @@ int tty_connect(void)
                         break;
 
                         default:
-                            tio_error_printf("Unknown outut mode");
+                            tio_error_printf("Unknown output mode");
                             exit(EXIT_FAILURE);
                         break;
                     }
 
                     /* Convert MSB to LSB bit order */
-                    if (option.map_o_msblsb)
+                    if (option.map_i_msb2lsb)
                     {
                         char ch = input_char;
                         input_char = 0;
@@ -2537,7 +2524,7 @@ int tty_connect(void)
                     }
 
                     /* Map input character */
-                    if ((input_char == '\n') && (option.map_i_nl_crnl) && (!option.map_o_msblsb))
+                    if ((input_char == '\n') && (option.map_i_nl_crnl) && (!option.map_i_msb2lsb))
                     {
                         printchar('\r');
                         printchar('\n');
@@ -2546,7 +2533,7 @@ int tty_connect(void)
                             do_timestamp = true;
                         }
                     }
-                    else if ((input_char == '\f') && (option.map_i_ff_escc) && (!option.map_o_msblsb))
+                    else if ((input_char == '\f') && (option.map_i_ff_escc) && (!option.map_i_msb2lsb))
                     {
                         printchar('\e');
                         printchar('c');
@@ -2628,152 +2615,17 @@ int tty_connect(void)
                                     switch (input_char)
                                     {
                                         case '\r': // Carriage return
-                                            if (line_length > 0)
-                                            {
-                                                // Save to history
-                                                if (history_count < MAX_HISTORY)
-                                                {
-                                                    history[history_count] = strndup(line, line_length);
-                                                    history_count++;
-                                                }
-                                                else
-                                            {
-                                                    free(history[0]);
-                                                    memmove(&history[0], &history[1], (MAX_HISTORY - 1) * sizeof(char*));
-                                                    history[MAX_HISTORY - 1] = strndup(line, line_length);
-                                                }
-                                            }
+                                            readline_input(input_char);
 
-                                            line[line_length] = '\0';
-                                            if (option.local_echo == false)
-                                            {
-                                                clear_line();
-                                            }
-                                            else
-                                            {
-                                                print("\r\n");
-                                            }
-
-                                            // Write line to tty device
-                                            tty_write(device_fd, line, line_length);
+                                            // Write current line to tty device
+                                            char *line = readline_get();
+                                            tty_write(device_fd, line, strlen(line));
                                             tty_sync(device_fd);
-
-                                            line_length = 0;
-                                            cursor_pos = 0;
-                                            history_index = history_count;
-                                            escape = 0;
-                                            break;
-
-                                        case 127: // Backspace
-                                            if (cursor_pos > 0)
-                                            {
-                                                memmove(&line[cursor_pos - 1], &line[cursor_pos], line_length - cursor_pos);
-                                                line_length--;
-                                                cursor_pos--;
-                                                line[line_length] = '\0';
-                                                print_line(line, cursor_pos);
-                                            }
-                                            forward = false;
-                                            escape = 0;
-                                            break;
-
-                                        case 27: // Escape
-                                            escape = 1;
-                                            forward = false;
-                                            break;
-
-                                        case '[':
-                                            if (escape == 1)
-                                            {
-                                                escape = 2;
-                                            }
-                                            else
-                                            {
-                                                escape = 0;
-                                            }
-                                            break;
-
-                                        case 'A':
-                                            if (escape == 2)
-                                            {
-                                                // Up arrow
-                                                if (history_index > 0)
-                                                {
-                                                    history_index--;
-                                                    strncpy(line, history[history_index], MAX_LINE_LEN-1);
-                                                    line_length = strlen(line);
-                                                    cursor_pos = line_length;
-                                                    print_line(line, cursor_pos);
-                                                }
-                                            }
-                                            forward = false;
-                                            escape = 0;
-                                            break;
-
-                                        case 'B':
-                                            if (escape == 2)
-                                            {
-                                                // Down arrow
-                                                if (history_index < history_count - 1)
-                                                {
-                                                    history_index++;
-                                                    strncpy(line, history[history_index], MAX_LINE_LEN-1);
-                                                    line_length = strlen(line);
-                                                    cursor_pos = line_length;
-                                                    print_line(line, cursor_pos);
-                                                }
-                                                else if (history_index == history_count - 1)
-                                                {
-                                                    history_index++;
-                                                    line_length = 0;
-                                                    cursor_pos = 0;
-                                                    line[line_length] = '\0';
-                                                    print_line(line, cursor_pos);
-                                                }
-                                            }
-                                            forward = false;
-                                            escape = 0;
-                                            break;
-
-                                        case 'C':
-                                            if (escape == 2)
-                                            {
-                                                // Right arrow
-                                                if (cursor_pos < line_length)
-                                                {
-                                                    cursor_pos++;
-                                                    print("\x1b[C");
-                                                }
-                                            }
-                                            forward = false;
-                                            escape = 0;
-                                            break;
-
-                                        case 'D':
-                                            if (escape == 2)
-                                            {
-                                                    // Left arrow
-                                                    if (cursor_pos > 0)
-                                                    {
-                                                        cursor_pos--;
-                                                        print("\b");
-                                                    }
-                                            }
-                                            forward = false;
-                                            escape = 0;
                                             break;
 
                                         default:
-                                            if (line_length < MAX_LINE_LEN - 1)
-                                            {
-                                                memmove(&line[cursor_pos + 1], &line[cursor_pos], line_length - cursor_pos);
-                                                line[cursor_pos] = input_char;
-                                                line_length++;
-                                                cursor_pos++;
-                                                line[line_length] = '\0';
-                                                print_line(line, cursor_pos);
-                                            }
-                                            escape = 0;
+                                            readline_input(input_char);
+                                            forward = false;
                                             break;
                                     }
 
